/*******************************************************************************
 * This file is part of QuickSched.
 * Coypright (c) 2013 Pedro Gonnet (pedro.gonnet@durham.ac.uk)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 ******************************************************************************/

/* Config parameters. */
#include "../config.h"

/* Standard includes. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Local includes. */
#include "error.h"
#include "cycle.h"
#include "atomic.h"
#include "lock.h"
#include "task.h"
#include "qsched.h"
#include "queue.h"
#include "res.h"

/**
 * @brief Compute an overlap score for a pair of tasks.
 * 
 * @param s The #qsched in which the tasks and their resources live.
 * @param a Index of the first #task or #qsched_task_none.
 * @param b Index of the second #task or #qsched_task_none.
 *
 * @return The Jaccard similarity of the resources of both tasks.
 */
 
float queue_task_overlap ( struct qsched *s , qsched_task_t a , qsched_task_t b ) {
  /* Trivial case, one of the tasks is NULL. */
  if (a == qsched_task_none || b == qsched_task_none) return 0.0f;
  
  /* Get a handle on the two tasks. */
  const struct task *ta = &s->tasks[a];
  const struct task *tb = &s->tasks[b];
  
  /* Collect the locks/uses resource ids for both tasks. */
  const int nr_res_a = ta->nr_locks + ta->nr_uses;
  const int nr_res_b = tb->nr_locks + tb->nr_uses;
  if (nr_res_a == 0 || nr_res_b == 0) return nr_res_a == nr_res_b ? 1.0f : 0.0f;
  struct res *res_a[nr_res_a], *res_b[nr_res_b];
  for (int k = 0; k < ta->nr_locks; k++)
    res_a[k] = &s->res[ta->locks[k]];
  for (int k = 0; k < ta->nr_uses; k++)
    res_a[ta->nr_locks + k] = &s->res[ta->uses[k]];
  for (int k = 0; k < tb->nr_locks; k++)
    res_b[k] = &s->res[tb->locks[k]];
  for (int k = 0; k < tb->nr_uses; k++)
    res_b[tb->nr_locks + k] = &s->res[tb->uses[k]];
  
  /* Compute the resource union, which is just the sum of the
     resource sizes. */
  int res_union = 0;
  for (int k = 0; k < nr_res_a; k++)
    res_union += res_a[k]->size;
  for (int k = 0; k < nr_res_b; k++)
    res_union += res_b[k]->size;
  if (res_union == 0) return 1.0f;
  
  /* Compute the resource intersection. */
  int res_isect = 0;
  for (int k = 0; k < nr_res_a; k++) {
    const struct res *ra = res_a[k];
    for (int j = 0; j < nr_res_b; j++) {
      const struct res *rb = res_b[j];
      if (ra->data <= rb->data && rb->data < ra->data + ra->size) {
        if (rb->data + rb->size < ra->data + ra->size)
          res_isect += rb->size;
        else
          res_isect += ra->data + ra->size - rb->data;
      } else if (rb->data <= ra->data && ra->data < rb->data + rb->size) {
        if (ra->data + ra->size < rb->data + rb->size)
          res_isect += ra->size;
        else
          res_isect += rb->data + rb->size - ra->data;
      }
    }
  }
  
  /* Return the Jaccard similarity, i.e. the ratio of the intersection
     and the union. */
  return ((float)res_isect) / (res_union - res_isect);
}


/**
 * @brief Get a task index from the given #queue.
 *
 * @param q The #queue.
 * @param s The #qsched in which this queue's tasks lives.
 * @param insist If set, wait at the queue's lock, otherwise fail.
 * @param last Index of the previous task executed, or -1.
 *
 * @return The task ID or -1 if no available task could be found.
 */
 
int queue_get ( struct queue *q , struct qsched *s , int insist , int last ) {

    const struct task *tasks = s->tasks;

    /* Should we even try? */
    if ( q->count == 0 )
        return -1;
        
    /* Lock this queue. */
    TIMER_TIC
    if ( insist ) {
        if ( lock_lock( &q->lock ) != 0 )
            error( "Failed to lock queue." );
        }
    else if ( lock_trylock( &q->lock ) != 0 )
        return qsched_task_none;
    TIMER_TOC( s , qsched_timer_qlock );
        
    /* Get a pointer to the indices. */
    int *inds = q->inds;
    int count = q->count;
    
    /* Data for the sliding window in which to try the task with the
       best overlap with the previous task. */
    struct {
      int ind, tid;
      float score;
    } window[queue_search_window];
    int window_count = 0;
    qsched_task_t tid = qsched_task_none;
    int ind = -1;
        
    /* Loop over the queue entries. */
    for (int k = 0 ; k < count ; k++ ) {
      if (k < queue_search_window) {
        window[window_count].ind = k;
        window[window_count].tid = inds[k];
        window[window_count].score = queue_task_overlap(s, last, inds[k]);
        window_count += 1;
      } else {
        int ind_max = 0;
        for (int i = 1; i < window_count; i++)
          if (window[i].score > window[ind_max].score) ind_max = i;
        if (qsched_locktask(s, window[ind_max].tid)) {
          tid = window[ind_max].tid;
          ind = window[ind_max].ind;
          // message("best task has overlap %f.", window[ind_max].score);
          break;
        } else {
          window[ind_max].ind = k;
          window[ind_max].tid = inds[k];
          window[ind_max].score = queue_task_overlap(s, last, inds[k]);
        }
      }
    }
    
    /* If we didn't get a task, loop through whatever is left in the window. */
    if (tid == qsched_task_none) {
      while (window_count > 0) {
        int ind_max = 0;
        for (int i = 1; i < window_count; i++)
          if (window[i].score > window[ind_max].score) ind_max = i;
        if (qsched_locktask(s, window[ind_max].tid)) {
          tid = window[ind_max].tid;
          ind = window[ind_max].ind;
          // message("best task has overlap %f.", window[ind_max].score);
          break;
        } else {
          window_count -= 1;
          window[ind_max] = window[window_count];
        }
      }
    }
    
    /* Did we get a task? */
    if ( tid != qsched_task_none ) {
    
        /* Swap the last element to the new heap position. */
        q->count = ( count -= 1 );
        inds[ind] = inds[ count ];
    
        /* Fix the heap. */
        int k = ind;
        long long int w = tasks[ inds[k] ].weight;
        if ( k > 0 && w > tasks[ inds[(k-1)/2] ].weight )
            while ( k > 0 ) {
                int j = (k - 1) / 2;
                if ( w > tasks[ inds[j] ].weight ) {
                    int temp = inds[j];
                    inds[j] = inds[k];
                    inds[k] = temp;
                    k = j;
                    }
                else
                    break;
                }
        else
            while ( 1 ) {
                int j = 2*k + 1;
                if ( j >= count )
                    break;
                if ( j+1 < count && tasks[ inds[j] ].weight < tasks[ inds[j+1] ].weight )
                    j = j + 1;
                if ( tasks[ inds[j] ].weight > w ) {
                    int temp = inds[j];
                    inds[j] = inds[k];
                    inds[k] = temp;
                    k = j;
                    }
                else
                    break;
                }
        
        } /* did we get a task? */
        
    /* Unlock the queue. */
    lock_unlock_blind( &q->lock );

    /* Return the task ID. */
    return tid;

    }


/**
 * @brief Add a task index to the given #queue.
 * 
 * @param q The #queue.
 * @param s The #qsched in which the tasks live.
 * @param tid The task index.
 */
 
void queue_put ( struct queue *q , struct qsched *s , int tid ) {

    int ind, j, temp;
    struct task *tasks = s->tasks;
    int *inds, *inds_new;
    
    /* Lock this queue. */
    if ( lock_lock( &q->lock ) != 0 )
        error( "Failed to lock queue." );
        
    /* Get a pointer to the indices. */
    inds = q->inds;
        
    /* Get the index of the new task. */
    ind = q->count;
    
    /* Does the queue need to be extended? */
    if ( ind >= q->size ) {
    
        /* Increase the queue size. */
        q->size *= queue_grow;
        
        /* Allocate the new indices. */
        if ( ( inds_new = (int *)malloc( sizeof(int) * q->size ) ) == NULL )
            error( "Failed to allocate new indices." );
            
        /* Copy the old indices. */
        memcpy( inds_new , inds , sizeof(int) * q->count );
        
        /* Clear the old indices and replace them with the new. */
        free( inds );
        q->inds = ( inds = inds_new );
    
        }
        
    /* Store the task index. */
    q->count += 1;
    inds[ind] = tid;
    
    /* Bubble up the new entry. */
    long long int w = tasks[ inds[ind] ].weight;
    while ( ind > 0 ) {
        j = (ind - 1)/2;
        if ( tasks[ inds[j] ].weight < w ) {
            temp = inds[j];
            inds[j] = inds[ind];
            inds[ind] = temp;
            ind = j;
            }
        else
            break;
        }
        
    /* Unlock the queue. */
    lock_unlock_blind( &q->lock );

    }


/**
 * @brief Clean up a queue and free its memory.
 */
 
void queue_free ( struct queue *q ) {

    /* Free the inds. */
    if ( q->inds != NULL )
        free( (void *)q->inds );

    }


/** 
 * @brief Initialize the given #queue.
 *
 * @param q The #queue.
 * @param size The maximum size of the queue.
 */
 
void queue_init ( struct queue *q , int size ) {
    
    /* Allocate the task list if needed. */
    if ( q->inds == NULL || q->size < size ) {
        if ( q->inds != NULL )
            free( (int *)q->inds );
        q->size = size;
        if ( ( q->inds = (int *)malloc( sizeof(int) * size ) ) == NULL )
            error( "Failed to allocate queue inds." );
        }
    q->size = size;
    
    /* Init the lock. */
    lock_init( &q->lock );
        
    /* Init the count. */
    q->count = 0;
    
    }

